package logic;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.regex.Pattern;

import logic.code.CodeItem;
import logic.code.CommentItem;
import logic.code.SingleLineItem;
import logic.code.StatementItem;


public class JavaParser {

	BufferedReader stream;
	
	private CodeItem item = null;
	private int lineNum;
	private String source = "";
	
	private static Pattern shortComment, startComment, inlineComment;
	private static Pattern classPat, method, switchPat; 
	private static Pattern loops, ifPat, tryPat;
	private static Pattern importPat, glPat;
	private static Pattern quotation; 
	
	static {
		// finds all comments
		shortComment 	= Pattern.compile("\\s*//.*");								// matches: // (at the beginning) (\\s == [ \t\n\x0B\f\r])
		startComment 	= Pattern.compile("\\s*/\\*.*");							// matches: /*
		inlineComment 	= Pattern.compile("//");									// matches: //
		
		// finds methods
		classPat		= Pattern.compile(".*class [A-Z,a-z,\\s]+.*");				// matches: string class string
		method			= Pattern.compile("\\s*(private|public|protected|static|synchronized| ){1,6}\\s+" + // vorher: ([A-Z,a-z, ,\\,<,>,\\[,\\]]+)?
						"[A-Z,a-z,0-9,<,>,\\[,\\]]+\\s+[A-Z,a-z,0-9]+\\s*\\(([A-Z,a-z,0-9, ,\\,<,>,\\[,\\]]+)?\\).*");	// matches: a method
		loops			= Pattern.compile("\\s*(while|for|do)[^a-zA-z].*");					// matches: for () | while() | do
		ifPat			= Pattern.compile("\\s*(if|else)[^a-zA-z].*");						// matches: if ()
		switchPat		= Pattern.compile("\\s*switch[^a-zA-z].*");							// matches: switch()
		tryPat			= Pattern.compile("\\s*(try|catch)[^a-zA-z].*");					// matches: try{ and catch()
		
		// find single line statements
		importPat		= Pattern.compile("[ \\t\\r\\n\\v\\f]*(import|package)\\s+.*");	// matches: import | package
		glPat			= Pattern.compile("\\s*(gl.gl).*");						// matches: gl statements
		// keywords for methods / statements
		quotation		= Pattern.compile("\"");									// matches: "
	}
	
	/**
	 * This class creates a code hierarchy from java source files
	 * 
	 * @param inputStream	a text file input stream
	 */
	public JavaParser(File file) {
		
		try {
			stream = getBufferedReader(file);
			item = parse();
			
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
		
		lineNum = 1;
	}
	
	public CodeItem getCodeItem() {
		
		
		while(item.getParent() != null)
			item = item.getParent();
		
		return item;
	}
	
	public String getSourceCode(){
		return source;
	}

	private BufferedReader getBufferedReader(File file) throws IOException{
		
		BufferedReader inputStream = null;
		
		try {
			inputStream = new BufferedReader(new FileReader(file));
			
		} catch (FileNotFoundException fnfe) {
			fnfe.printStackTrace();
		}
		
		return inputStream;
		
	}
	
	private CodeItem parse() throws IOException{
		String fileStream = "";
		String line;
		while( (line = stream.readLine()) != null) {
			source += line +"\n";
			parseLine(line);
			lineNum++;
			fileStream += "\n"+line;
		}
		
		return item;
	}
	
	private void parseLine(String line) {
		
		// Convert tabulators to spaces
		line = line.replace("\t", "     ");
		
		if (item == null)
			item = new CodeItem(null, lineNum);
		
		// if comment: skip rest...
		findComment(line);
		
		if (item.getMode() != CodeItem.COMMENT) {
			
			// don't consider comments at the end of a line
			String[] spLine = inlineComment.split(line);
			// don't consider quoted elements
			spLine = quotation.split(spLine[0]);
			String parseLine = "";
			
			for (int i = 0; i < spLine.length; i++) {
				if (i%2 == 0)
					parseLine = parseLine.concat(spLine[i]);
			}
				
			findClass(parseLine);
			findMethod(parseLine);
			findLoops(parseLine);
			findIfStatements(parseLine);
			findSwitch(parseLine);
			findTryCatch(parseLine);
			findSingleLineStatements(parseLine);
			findGLStatements(parseLine);
		}
		
		item.addLine(line, lineNum);
		item = item.getCurrentItem();
	}
	
	private void findComment(String line) {
		
		boolean startCom = false, shortCom = false;
		
		shortCom = shortComment.matcher(line).matches();
		startCom = startComment.matcher(line).matches();
				
		if (shortCom || startCom) {
			item = new CommentItem(item, lineNum, shortCom);
		}
	
	}
	
	private void findClass(String line) {
		
		boolean startObj = classPat.matcher(line).matches();
		
		if (startObj) {
			item = new CodeItem(item, lineNum, CodeItem.CLASS);
		}
	}
	
	private void findMethod(String line) {
		
		boolean startObj = method.matcher(line).matches();
		
		// otherwise statements such as return something(); match and else if (bla)
		if (line.contains(";"))
			startObj = false;
		else if (line.contains("else if"))
			startObj = false;
		
		if (startObj) {
			item = new CodeItem(item, lineNum, CodeItem.METHOD);
		}
	}
	
	private void findLoops(String line) {
		
		boolean startObj = loops.matcher(line).matches();
				
		if (startObj) {
			item = new StatementItem(item, lineNum);
		}
	}
	
	private void findIfStatements(String line) {
		
		boolean startObj = ifPat.matcher(line).matches();
		
		if (startObj) {
			item = new StatementItem(item, lineNum, CodeItem.IF);
		}
	}
	
	private void findTryCatch(String line) {
		
		boolean startObj = tryPat.matcher(line).matches();
		
		if (startObj) {
			item = new StatementItem(item, lineNum, CodeItem.TRY);
		}
	}
	
	private void findSwitch(String line) {
		
		boolean startObj = switchPat.matcher(line).matches();
		
		if (startObj) {
			item = new StatementItem(item, lineNum, CodeItem.SWITCH);
		}
	}
	
	private void findSingleLineStatements(String line) {
		
		boolean startObj = importPat.matcher(line).matches();
		
		if (startObj) {
			item = new SingleLineItem(item, lineNum);
		}
	}
	
	private void findGLStatements(String line) {
		
		boolean startObj = glPat.matcher(line).matches();
		
		if (startObj) {
			item = new SingleLineItem(item, lineNum, CodeItem.GL_MODE);
		}
	}
	
}
