package logic.code;

import java.awt.Color;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class CodeItem {
	
	public static final int UNDEFINED = 0, PACKAGE = 1, IMPORT = 2, COMMENT = 3, 
							CLASS = 4, METHOD = 5, LOOP = 6, SWITCH = 7, 
							IF = 8, TRY = 9, GL_MODE = 10;
							
	public static final float Z_OFFSET = 0.01f;
	
	protected static Pattern whiteSpacePat; 	
	
	
	protected CodeItem parent;
	protected LinkedList<CodeItem> children;
	
	protected Color color = new Color(0, 0, 0);
	protected LinkedList<String> text;
	
	protected int firstLine, lastLine, maxLength, lineStart;
	protected int brackets;
	protected int mode;
	protected String name;
	protected CodeItem currentItem;
	protected int curLine = 0, curChild = 0, curTextLine = 0;
	protected float alpha = 1.0f;
	protected float zIndex = 0.0f;
	
	static {
		whiteSpacePat = Pattern.compile("\\s*");
	}
			
	public CodeItem(CodeItem parent, int lineNumber) {

		this.parent = parent;
		
		if (parent != null)
			parent.addChild(this);
		
		children = new LinkedList<CodeItem>();
		text = new LinkedList<String>();
		currentItem = this;
		
		firstLine = lineNumber;
		lastLine = lineNumber;
		maxLength = 0;
		lineStart = Integer.MAX_VALUE;
		
		this.mode = UNDEFINED;
		name = "Undefined";
		brackets = -1;
		
		setZIndex();
	}

	public CodeItem(CodeItem parent, int lineNumber, int mode) {

		this.parent = parent;
		
		if (parent != null)
			parent.addChild(this);
		
		children = new LinkedList<CodeItem>();
		text = new LinkedList<String>();
		currentItem = this;
		
		firstLine = lineNumber;
		lastLine = lineNumber;
		maxLength = 0;
		lineStart = Integer.MAX_VALUE;
		
		this.mode = mode;
		brackets = -1;
		
		setMode();
		setZIndex();
	}
	
	private void setZIndex() {
		if (parent != null)
			zIndex = parent.zIndex+Z_OFFSET;
	}
	
	public float getZIndex() {
		return zIndex;
	}
	
	protected void setMode() {
		
		switch (mode) {
		case UNDEFINED:
			color = new Color(UNDEFINED, UNDEFINED, UNDEFINED);
			name = "Undefined";
			break;
		case COMMENT:
			color = new Color(COMMENT, COMMENT, COMMENT);
			name = "Comment";
			break;
		case CLASS:
			color = new Color(CLASS, CLASS, CLASS);
			name = "Class";
			break;
		case METHOD:
			color = new Color(METHOD, METHOD, METHOD);
			name = "Method";
			break;
		case LOOP:
			color = new Color(LOOP, LOOP, LOOP);
			name = "Loop";
			break;
		case IF:
			color = new Color(IF, IF, IF);
			name = "If Statement";
			break;
		case SWITCH:
			color = new Color(SWITCH, SWITCH, SWITCH);
			name = "Switch";
			break;
		case IMPORT:
			color = new Color(IMPORT, IMPORT, IMPORT);
			name = "Import";
			break;
		case PACKAGE:
			color = new Color(PACKAGE, PACKAGE, PACKAGE);
			name = "Package";
			break;
		case TRY:
			color = new Color(TRY, TRY, TRY);
			name = "Try";
			break;
		case GL_MODE:
			color = new Color(GL_MODE, GL_MODE, GL_MODE);
			name = "GL";
			break;
		default:
			color = new Color(UNDEFINED, UNDEFINED, UNDEFINED);
			name = "Undefined";
			break;
			
		}
	}
		
	public CodeItem getParent() {
		return parent;
	}
	
	public void addChild(CodeItem item) {
		children.add(item);
	}
	
	public LinkedList<CodeItem> getChildren() {
		return children;
	}
	
	public void addLine(String line, int lineNum) {
		
		text.add(line);
		lastLine = lineNum;
		
		if (parent != null)
			parent.setLastLine(lastLine);
		
		if (line.length() > maxLength) {
			maxLength = line.length();
			parent.setMaxLength(maxLength);
		}

		// ignore empty lines
		if (line.trim().length() > 0 || lineStart == Integer.MAX_VALUE) {
		
			Matcher matcher = whiteSpacePat.matcher(line);
			
			int whiteSpace = 0;
			if (matcher.find())
				whiteSpace = matcher.end();
			
			if (whiteSpace < lineStart)
				lineStart = whiteSpace;
		}
		
		findTerminator(line);
	}
	
	protected void setMaxLength(int maxLength) {
		
		maxLength++;		// this looks much better...
		
		if (this.maxLength < maxLength) {
			this.maxLength = maxLength;
			
			if (parent != null)
				parent.setMaxLength(maxLength);
		}
	}

	protected void setLastLine(int lastLine) {
		this.lastLine = lastLine;
		
		if (parent != null)
			parent.setLastLine(lastLine);
	}
	
	protected void findTerminator(String line) {
		boolean ob = line.contains("{");
		boolean cb = line.contains("}");
				
		if (brackets == -1 && ob)
			brackets = 0;
		
		if (ob)
			brackets++;
		if (cb)
			brackets--;
		
		if (brackets == 0)
			terminateItem();
	}
	
	protected void terminateItem() {
		currentItem = parent;
	}
	
	
	public CodeItem getCurrentItem() {
		return currentItem;
	}
	
	public int getMode() {
		return mode;
	}
	
	// reading...
	public String nextLine() {
		
		String nextLine = null;
		
		if (!children.isEmpty() && children.size() > curChild) {
			if(children.get(curChild).lastLine-firstLine < curLine)
				curChild++;
			
			if(children.size() > curChild && children.get(curChild).firstLine-firstLine <= curLine)
				nextLine = children.get(curChild).nextLine();	
		}
		
		if (nextLine == null && text.size() > curTextLine) {
			nextLine = text.get(curTextLine);
			curTextLine++;			
		}
		
		if (!this.hasNext())
			this.reset();
		
		curLine++;
		
		return nextLine;
	}
	
	public String getFirstLine() {
		curLine = 0;
		curChild = 0;
		curTextLine = 0;
		return nextLine();
	}
	
	public void reset() {
		curLine = 0;
		curChild = 0;
		curTextLine = 0;
		
		// reset children recursively
		Iterator<CodeItem> iter = children.iterator();
		
		while(iter.hasNext())
			iter.next().reset();
	}
	
	public boolean hasNext() {
		return (curLine <= lastLine) ? true : false;
	}
	
	public int firstLine() {
		return firstLine;
	}
	
	public int lastLine() {
		return lastLine;
	}
	
	protected int getCurLineCnt() {
		return curLine;
	}
	
	public int lineStart() {
		return lineStart;
	}
	
	public int maxLine() {
		return maxLength;
	}

	public Color getColor() {
		return color;
	}
	
	public LinkedList<CodeItem> toList() {
		
		LinkedList<CodeItem> allItems = new LinkedList<CodeItem>();
		allItems.add(this);
		
		Iterator<CodeItem> childIter = children.iterator();
		CodeItem curChild;
		
		while (childIter.hasNext()) {
			curChild = childIter.next();
			allItems.addAll(curChild.toList());
			
		}
		
		return allItems;
	}
	
	public void setAlpha(float alpha) {
		this.alpha = alpha;
	}
	
	public float getAlpha() {
		return alpha;
	}
}
