package opengl;

import gui.StatementPanel;

import java.awt.Color;
import java.nio.FloatBuffer;
import java.util.Iterator;
import java.util.LinkedList;

import javax.media.opengl.GL;
import javax.media.opengl.glu.GLU;

import opengl.common.FadeEffect;
import opengl.common.Shader;

import logic.code.CodeItem;

public class RectangleRenderer {
	
	private static final int T_COLORS = 0;
	
	// shader
	private static Shader rectangleShader;
	private static int[] tcolor = new int[1];
	private static boolean transferChanged = true;
	private static LinkedList<FadeEffect> fadingRects = new LinkedList<FadeEffect>();
	private static LinkedList<Integer> lucentRects = new LinkedList<Integer>();
	
	// data
	private static int[] rectList;
	private static float gradient1 = 1.0f, gradient2 = 0.3f;
	private static float highlightAlpha = 1.0f, globalAlpha = 1.0f;
	
	// openGL
	private static GLU glu = new GLU();
	
	public static void setDisplayListSize(int size) {
		rectList = new int[size];
	}
	
	
	public static void init(GL gl) {
		
		rectangleShader = new Shader(gl, "shader/rectangle-shader.vsh", "shader/rectangle-shader.fsh");

		int program = rectangleShader.getProgram();
		gl.glLinkProgramARB(program);
		gl.glUseProgram(program);
		
		gl.glUniform1iARB(gl.glGetUniformLocationARB(program, "transfer_tex"), 0);
		gl.glUniform1fARB(gl.glGetUniformLocationARB(program, "highlight_alpha"), highlightAlpha);
		gl.glUniform1fARB(gl.glGetUniformLocationARB(program, "global_alpha"), globalAlpha);
		gl.glValidateProgramARB(program);
        
        rectangleShader.checkLogInfo(gl, program);
        gl.glUseProgramObjectARB(0);
	}
	
	public static void draw(GL gl, int i) {

		generateTColors(gl);
		
		int program = rectangleShader.getProgram();
		gl.glUseProgram(program);
		
		gl.glUniform1fARB(gl.glGetUniformLocationARB(program, "highlight_alpha"), highlightAlpha);
		gl.glUniform1fARB(gl.glGetUniformLocationARB(program, "global_alpha"), globalAlpha);
		gl.glCallList(rectList[i]);
		
		gl.glUseProgram(0);
		
		// See errors...
		int glError = GL.GL_NO_ERROR;
    	glError = gl.glGetError();
    	if (glError != GL.GL_NO_ERROR)
    		System.out.println("Error while drawing Rectangles: " + glu.gluErrorString(glError));
	}
	
	public static void drawSelectedRectangle(GL gl, CodeItem item) {
		
		int program = rectangleShader.getProgram();
		gl.glUseProgram(program);
		
		gl.glUniform1fARB(gl.glGetUniformLocationARB(program, "highlight_alpha"), 
				1.0f+CodeRenderer.FADE_ALPHA-highlightAlpha);
		gl.glUniform1fARB(gl.glGetUniformLocationARB(program, "global_alpha"), 1.0f);
		drawRectangle(gl, item, true);
		gl.glUseProgram(0);
		
		// See errors...
		int glError = GL.GL_NO_ERROR;
    	glError = gl.glGetError();
    	if (glError != GL.GL_NO_ERROR)
    		System.out.println("Error while drawing Rectangles: " + glu.gluErrorString(glError));
	}
	
	
	public static void createList(GL gl, CodeItem item, int which) {
		
		Iterator<CodeItem> itemIter = item.toList().iterator();
		
		rectList[which] = gl.glGenLists(1);
		gl.glNewList(rectList[which], GL.GL_COMPILE);
		
		while (itemIter.hasNext())
			drawRectangle(gl, itemIter.next(), false);
		
		gl.glEndList();
		
		int glError = GL.GL_NO_ERROR;
    	glError = gl.glGetError();
    	if (glError != GL.GL_NO_ERROR)
    		System.out.println("Error while creating Rectangles: " + glu.gluErrorString(glError));
	}
	
	// Open GL...
	private static void drawRectangle(GL gl, CodeItem item, boolean selected) {
		
		Color col = item.getColor();
		float x1 = item.lineStart()*FontRenderer.CHAR_TRANS-FontRenderer.CHAR_TRANS*0.5f;
		float x2 = item.maxLine()*FontRenderer.CHAR_TRANS+FontRenderer.CHAR_TRANS*0.5f;
		float y1 = (item.firstLine()-0.8f+(FontRenderer.LINE_SPACING * .5f))*FontRenderer.LINE_SPACING;
		float y2 = (item.lastLine()+0.2f+(FontRenderer.LINE_SPACING * .5f))*FontRenderer.LINE_SPACING;
		float shadowID = (float)col.getBlue()/255.0f;
		float shadowAlpha = 0.4f;
		
		if (selected) {
			shadowID = 0.6f;
			shadowAlpha = 1.0f;
		}
		
		gl.glEnable(GL.GL_BLEND); // enable color and texture blending
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
		
		// draw shadow
		gl.glPushMatrix();
		gl.glTranslatef(0.2f, -0.2f, 0.0f);
		
		gl.glBegin(GL.GL_QUADS);
		
		gl.glColor4f(0.0f, 0.0f, shadowID, shadowAlpha);	// 0.11f -> skip color look-up
		gl.glVertex3f(x1, y1, -0.1f);      // Vertex Coord (Bottom Left)
		gl.glColor4f(0.0f, 0.0f, shadowID, shadowAlpha);
		gl.glVertex3f(x2, y1, -0.1f);      // Vertex Coord (Bottom Right)
		gl.glColor4f(0.0f, 0.0f, shadowID, shadowAlpha);
		gl.glVertex3f(x2, y2, -0.1f);      // Vertex Coord (Top Right)
		gl.glColor4f(0.0f, 0.0f, shadowID, shadowAlpha);
		gl.glVertex3f(x1, y2, -0.1f);      // Vertex Coord (Top Left)
		
		
		gl.glEnd();
		gl.glPopMatrix();
		
		// draw colored rectangle
		gl.glBegin(GL.GL_QUADS);
		
		// red:		look-up
		// green:	brightness
		// blue:	shadow
		// alpha:	normal
		gl.glColor4f(col.getRed()/255.0f, gradient1, 1.0f, 1.0f);
		gl.glVertex3f(x1, y1, -0.1f);      // Vertex Coord (Bottom Left)
		gl.glColor4f(col.getRed()/255.0f, gradient1, 1.0f, 1.0f);
		gl.glVertex3f(x2, y1, -0.1f);      // Vertex Coord (Bottom Right)
		gl.glColor4f(col.getRed()/255.0f, gradient2, 1.0f, 1.0f);
		gl.glVertex3f(x2, y2, -0.1f);      // Vertex Coord (Top Right)
		gl.glColor4f(col.getRed()/255.0f, gradient2, 1.0f, 1.0f);
		gl.glVertex3f(x1, y2, -0.1f);      // Vertex Coord (Top Left)
		
		
		gl.glEnd();
		
		gl.glDisable(GL.GL_BLEND);
	}

	private static void generateTColors(GL gl) {
		
		if (!transferChanged)
			return;
		
		Color[] rawMapping = StatementPanel.getColorMapping();
		rawMapping = fadeAnimation(rawMapping);
		
		FloatBuffer tColors = convertMapping(rawMapping);
		
    	gl.glEnable(GL.GL_TEXTURE_1D);
    	gl.glActiveTexture(GL.GL_TEXTURE0);
		gl.glGenTextures(1, tcolor, T_COLORS);
		gl.glBindTexture(GL.GL_TEXTURE_1D, tcolor[T_COLORS]);

		gl.glTexEnvf(GL.GL_TEXTURE_ENV, GL.GL_TEXTURE_ENV_MODE, GL.GL_REPLACE);
		gl.glTexParameteri(GL.GL_TEXTURE_1D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST);
		gl.glTexParameteri(GL.GL_TEXTURE_1D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST);
    	gl.glTexImage1D(GL.GL_TEXTURE_1D, 0, GL.GL_RGBA, tColors.capacity()/4, 0, GL.GL_RGBA, GL.GL_FLOAT, tColors);
    	
    	gl.glDisable(GL.GL_TEXTURE_1D);
		
    	if (fadingRects.isEmpty())
    		transferChanged = false;
	}
	
	private static Color[] fadeAnimation(Color[] colors) {

		Color col1, col2;
		int alpha;
		
		// keep lucent rects lucent
		Iterator<Integer> iterLucent = lucentRects.iterator();
		int itemID = 0;
		
		while (iterLucent.hasNext()) {
			itemID = iterLucent.next(); 
			col1 = colors[itemID];
			col2 = new Color(col1.getRed(), col1.getGreen(), col1.getBlue(), 0);
			colors[itemID] = col2;
		}

		// nothing else to do here...
		if (fadingRects.isEmpty())
			return colors;
		
		LinkedList<FadeEffect> removeList = new LinkedList<FadeEffect>();
		Iterator<FadeEffect> iter = fadingRects.iterator();
		FadeEffect effect;
		
		while (iter.hasNext()) {
			effect = iter.next();
			col1 = colors[effect.getItemID()];
			alpha = effect.getAlpha();
			col2 = new Color(col1.getRed(), col1.getGreen(), col1.getBlue(), alpha);
			colors[effect.getItemID()] = col2;
			
			if (alpha == 0) {
				removeList.add(effect);
				lucentRects.add(effect.getItemID());
			}
			else if (alpha == 255)
				removeList.add(effect);
			
			if (effect.isFadeIn())
				lucentRects.remove((Integer)effect.getItemID());
		}
		
		fadingRects.removeAll(removeList);
		
		return colors;
	}
	
	private static FloatBuffer convertMapping(Color[] map) {
			
		float[] transfer = new float[map.length*4];
		
		for(int i = 0; i < map.length; i++) {
			
			transfer[i*4] = (float)map[i].getRed()/255.0f;
			transfer[i*4+1] = (float)map[i].getGreen()/255.0f;
			transfer[i*4+2] = (float)map[i].getBlue()/255.0f;
			transfer[i*4+3] = (float)map[i].getAlpha()/255.0f;
			
		}
				
		return FloatBuffer.wrap(transfer);
	}
	
	public static void animateStatement(int which, boolean show) {
		fadingRects.add(new FadeEffect(which, show));
		transferChanged = true;
	}
	
	public static void colorsChanged(boolean changed) {
		transferChanged = changed;
	}
	
	
	public static void setGlobalAlpha(float alpha) {
		globalAlpha = (float)alpha/100.0f;
	}
	
	public static void setHighlightAlpha(float alpha) {
		highlightAlpha = alpha;
	}

}
