package opengl.common;

import java.awt.image.BufferedImage;
import java.awt.image.PixelGrabber;
import java.awt.image.Raster;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;

import javax.imageio.ImageIO;

import com.sun.opengl.util.BufferUtil;

public class ImageFileReader {

	public static ByteBuffer readTexture(File file) throws IOException {
		
		return readTexture(file.getAbsolutePath(), false);
	}
	
	public static ByteBuffer readTexture(String filename) throws IOException {
		
		return readTexture(filename, false);
	}
	
	public static ByteBuffer readTexture(String filename, boolean alphaChannel) throws IOException{
		
		BufferedImage image = readImage(filename);
		return readData(image, alphaChannel);// new ByteBuffer();
	}
	
	public static BufferedImage readImage(File file) throws IOException{
		return readImage(file.getAbsolutePath());
	}
	

	public static BufferedImage readImage(String filename) throws IOException{

		return ImageIO.read(getResourceAsStream(filename));
	}

	public static BufferedReader getResourceAsReader(final String filename) throws IOException {
		
		return new BufferedReader(new InputStreamReader(getResourceAsStream(filename)));
	}
	
    public static InputStream getResourceAsStream(final String filename) throws IOException {
     
    	// Try to load resource from jar
        String convertedFileName = filename.replace('\\', '/');
        InputStream stream = ImageFileReader.class.getClassLoader().getResourceAsStream(convertedFileName);
        // If not found in jar, then load from disk
        if (stream == null) {
            return new FileInputStream(convertedFileName);
        } else {
            return stream;
        }
    }
    
    public static ByteBuffer readData(BufferedImage image, boolean alphaChannel) {
        int[] packedPixels = new int[image.getWidth() * image.getHeight()];

        PixelGrabber pixelgrabber = new PixelGrabber(image, 0, 0, image.getWidth(),
        		image.getHeight(), packedPixels, 0, image.getWidth());
        try {
            pixelgrabber.grabPixels();
        } catch (InterruptedException e) {
            throw new RuntimeException();
        }

        int bytesPerPixel = alphaChannel ? 4 : 3;
        ByteBuffer unpackedPixels = BufferUtil.newByteBuffer(packedPixels.length * bytesPerPixel);

        for (int row = image.getHeight() - 1; row >= 0; row--) {
            for (int col = 0; col < image.getWidth(); col++) {
                int packedPixel = packedPixels[row * image.getWidth() + col];
                unpackedPixels.put((byte) ((packedPixel >> 16) & 0xFF));
                unpackedPixels.put((byte) ((packedPixel >> 8) & 0xFF));
                unpackedPixels.put((byte) ((packedPixel >> 0) & 0xFF));
                if (alphaChannel) {
                    unpackedPixels.put((byte) ((packedPixel >> 24) & 0xFF));
                }
            }
        }

        unpackedPixels.flip();
    	
    	return unpackedPixels;
    }
    
    public static ByteBuffer readGrayscalePixels(BufferedImage img) {
        
    	int[] packedPixels = new int[img.getWidth() * img.getHeight()];

        PixelGrabber pixelgrabber = new PixelGrabber(img, 0, 0, img.getWidth(), img.getHeight(), packedPixels, 0, img.getWidth());
        try {
            pixelgrabber.grabPixels();
        } catch (InterruptedException e) {
            throw new RuntimeException();
        }

        
        ByteBuffer unpackedPixels = BufferUtil.newByteBuffer(packedPixels.length);

        for (int row = img.getHeight() - 1; row >= 0; row--) {
            for (int col = 0; col < img.getWidth(); col++) {
                int packedPixel = packedPixels[row * img.getWidth() + col];
                unpackedPixels.put((byte) ((packedPixel >> 16) & 0xFF));
            }
        }

        unpackedPixels.flip();


        return unpackedPixels;
    }

    
    public static FloatBuffer readGrayscaleData(BufferedImage img, float maxVal) {
    	
    	Raster raster = img.getRaster();
    	
    	float[] data = new float[img.getHeight()*img.getWidth()*3];
		data = raster.getPixels(0, 0, img.getWidth(), img.getHeight(), data);

		// normalize
		for (int i = 0; i < data.length; i++) {
			
			data[i] = data[i]/maxVal;
		}
		
		FloatBuffer dataBuffer = FloatBuffer.wrap(data);
		
    	
    	return dataBuffer;
    }
    
    public static FloatBuffer readRGBData(BufferedImage img, float maxVal) {
    	
    	Raster raster = img.getRaster();
    	
    	float[] data = new float[img.getHeight()*img.getWidth()*3];
		data = raster.getPixels(0, 0, img.getWidth(), img.getHeight(), data);

		// normalize
		for (int i = 0; i < data.length; i++) {
			
			data[i] = data[i]/maxVal;
			
			if (data[i] > 1.0f || data[i] < 0.0f) {
				System.out.println("Data out of bounds: " + data[i]);
			}
		}
		
		FloatBuffer dataBuffer = FloatBuffer.wrap(data);
    	
    	return dataBuffer;
    }
    

}
